/*******************************************************************************
 * Copyright (c) 2008, 2010 Xuggle Inc.  All rights reserved.
 *  
 * This file is part of Xuggle-Xuggler-Red5.
 *
 * Xuggle-Xuggler-Red5 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Xuggle-Xuggler-Red5 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Xuggle-Xuggler-Red5.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package com.xuggle.red5;

import org.apache.mina.core.buffer.IoBuffer;
import org.red5.server.api.stream.IBroadcastStream;
import org.red5.server.api.stream.IStreamListener;
import org.red5.server.api.stream.IStreamPacket;
import org.red5.server.net.rtmp.event.AudioData;
import org.red5.server.net.rtmp.event.IRTMPEvent;
import org.red5.server.net.rtmp.event.Notify;
import org.red5.server.net.rtmp.event.VideoData;

import com.xuggle.xuggler.IAudioResampler;
import com.xuggle.xuggler.IAudioSamples;
import com.xuggle.xuggler.ICodec;
import com.xuggle.xuggler.IContainer;
import com.xuggle.xuggler.IContainerFormat;
import com.xuggle.xuggler.IVideoPicture;
import com.xuggle.xuggler.IPacket;
import com.xuggle.xuggler.IRational;
import com.xuggle.xuggler.ISimpleMediaFile;
import com.xuggle.xuggler.IStream;
import com.xuggle.xuggler.IStreamCoder;
import com.xuggle.xuggler.IVideoResampler;
import com.xuggle.xuggler.SimpleMediaFile;

import com.xuggle.red5.io.BroadcastStream;
import com.xuggle.red5.io.IRTMPEventIOHandler;
import com.xuggle.red5.io.Red5HandlerFactory;
import com.xuggle.red5.io.Red5Message;
import com.xuggle.red5.io.Red5StreamingQueue;

/**
 * Transcodes video and audio of a particular type from a red5 stream
 * into a new audio and video red5 stream.
 * <p>
 * It does this by opening an input URL, decoding all packets
 * in the container, resampling the decoded data if needed,
 * and then encoding into a new container.
 * </p>
 * <p>
 * It also allows hooking of call-back
 * listeners so you can be notified of key event and
 * potentially modify decoded data before re-encoding.
 * </p>
 */
public class Transcoder implements Runnable
{
  // This line initializes the AAFFMPEG IO libraries and gets a factory
  // we can register streams with.
  final static private Red5HandlerFactory mFactory = Red5HandlerFactory.getFactory();

  private final IBroadcastStream mInputStream;
  private final Red5StreamingQueue mInputQueue;
  private final IStreamListener mInputListener;
  private final BroadcastStream mOutputStream;
  private final ISimpleMediaFile mOutputInfo;
  private final IRTMPEventIOHandler mOutputHandler;
  private final IPacketListener mPacketListener;
  private final IVideoPictureListener mVideoPictureListener;
  private final IAudioSamplesListener mSamplesListener;

  private volatile boolean mIsRunning=false;
  private volatile boolean mKeepRunning=true;

  private String mInputURL;
  private String mOutputURL;
  private IContainer mOutContainer;
  private IStreamCoder mOutAudioCoder;
  private IStreamCoder mOutVideoCoder;
  private IContainer mInContainer;
  private IStreamCoder mInAudioCoder;
  private IStreamCoder mInVideoCoder;
  private int mAudioStreamId;
  private IAudioResampler mAudioResampler;
  private IVideoResampler mVideoResampler;
  private int mVideoStreamId;

  /**
   * Create a new transcoder object.
   * 
   * All listeners are set to null.
   *  
   * @param aInputStream The stream to get input packets from.
   * @param aOutputStream The stream to publish output packets to.
   * @param aOutputInfo Meta data about what type of packets you want to publish.
   */
  public Transcoder(
      IBroadcastStream aInputStream,
      BroadcastStream aOutputStream,
      ISimpleMediaFile aOutputInfo
  )
  {
    this(aInputStream, aOutputStream, aOutputInfo, null, null, null);
  }
  
  /**
   * Create a new transcoder object.
   * 
   * All listeners are set to null.
   *  
   * @param aInputStream The stream to get input packets from.
   * @param aOutputStream The stream to publish output packets to.
   * @param aOutputInfo Meta data about what type of packets you want to publish.
   * @param aPacketListener A packet listener that will be called for interesting events.  Or null to disable.
   * @param aSamplesListener A Audio Samples listener that will be called for interesting events.  Or null to disable.
   * @param aPictureListener A Video Picture listener that will be called for interesting events.  Or null to disable.
   */
  public Transcoder(
      IBroadcastStream aInputStream,
      BroadcastStream aOutputStream,
      ISimpleMediaFile aOutputInfo,
      IPacketListener aPacketListener,
      IAudioSamplesListener aSamplesListener,
      IVideoPictureListener aPictureListener
  )
  {
    if (aInputStream==null)
      throw new IllegalArgumentException("must pass input stream");
    if (aOutputStream == null)
      throw new IllegalArgumentException("must pass output stream");
    if (aOutputInfo == null)
      throw new IllegalArgumentException("must pass output stream info");
    
    mInputStream = aInputStream;
    mOutputStream = aOutputStream;
    mOutputInfo = aOutputInfo;
    mInputQueue = new Red5StreamingQueue();
    mPacketListener = aPacketListener;
    mSamplesListener = aSamplesListener;
    mVideoPictureListener = aPictureListener;
    
    mAudioStreamId = -1;
    mVideoStreamId = -1;

    // Check that we have valid input and output formats if specified
    if (mOutputInfo.getContainerFormat() != null)
    {
      IContainerFormat fmt = mOutputInfo.getContainerFormat();
      if (fmt.getInputFormatShortName() != "flv")
        throw new IllegalArgumentException("currently we only support inputs from FLV files");
    }
    // Make sure if specifying audio that we have all required parameters set.
    if (mOutputInfo.hasAudio())
    {
      if (!mOutputInfo.isAudioBitRateKnown() || mOutputInfo.getAudioBitRate() <= 0)
        throw new IllegalArgumentException("must set audio bit rate when outputting audio");
      if (!mOutputInfo.isAudioChannelsKnown() || mOutputInfo.getAudioChannels() <= 0)
        throw new IllegalArgumentException("must set audio channels when outputting audio");
      if (!mOutputInfo.isAudioSampleRateKnown() || mOutputInfo.getAudioSampleRate() <= 0)
        throw new IllegalArgumentException("must set audio sample rate when outputting audio");
      if (mOutputInfo.getAudioCodec() == ICodec.ID.CODEC_ID_NONE)
        throw new IllegalArgumentException("must set audio code when outputting audio");
    }
    if (mOutputInfo.hasVideo())
    {
      if (!mOutputInfo.isVideoBitRateKnown() || mOutputInfo.getVideoBitRate() <= 0)
        throw new IllegalArgumentException("must set video bit rate when outputting video");
      if (!mOutputInfo.isVideoHeightKnown() || mOutputInfo.getVideoHeight() <= 0)
        throw new IllegalArgumentException("must set video height when outputting video");
      if (!mOutputInfo.isVideoWidthKnown() || mOutputInfo.getVideoWidth() <= 0)
        throw new IllegalArgumentException("must set video width when outputting video");
      if (mOutputInfo.getVideoCodec() == ICodec.ID.CODEC_ID_NONE)
        throw new IllegalArgumentException("must set video codec when outputting video");
    }
    if (!(mOutputInfo.hasAudio() || mOutputInfo.hasVideo()))
      throw new IllegalArgumentException("must output either audio or video");
    
    // Make FFMPEG return back if a read takes longer than 2 seconds.  Unfortunately
    // it means FFMPEG will think the stream has ended, but them's the breaks.
    //mInputQueue.setReadTimeout(new TimeValue(2, TimeUnit.SECONDS));
    mInputListener = new IStreamListener(){
      public void packetReceived(IBroadcastStream aStream, IStreamPacket aPacket)
      {
        try {

          IoBuffer buf = aPacket.getData();
          if (buf != null)
            buf.rewind();
          if (buf==null || buf.remaining()==0)
          {
            return;
          }
          
          if (aPacket instanceof AudioData)
          {
            mInputQueue.put(new Red5Message(Red5Message.Type.AUDIO, (AudioData)aPacket));
          } else if (aPacket instanceof VideoData)
          {
            Red5Message.Type type = Red5Message.Type.INTERFRAME;
            VideoData dataPacket = (VideoData)aPacket;
            switch(dataPacket.getFrameType())
            {
            case DISPOSABLE_INTERFRAME:
              type = Red5Message.Type.DISPOSABLE_INTERFRAME;
              break;
            case INTERFRAME:
              type = Red5Message.Type.INTERFRAME;
              break;
            case KEYFRAME:
            case UNKNOWN:
              type = Red5Message.Type.KEY_FRAME;
              break;
            }
            if (type != Red5Message.Type.DISPOSABLE_INTERFRAME) // The FFMPEG FLV decoder doesn't handle disposable frames
            {
              mInputQueue.put(new Red5Message(type, dataPacket));
            }
          } else if (aPacket instanceof IRTMPEvent)
          {
            Red5Message.Type type = Red5Message.Type.OTHER;
            IRTMPEvent dataPacket = (IRTMPEvent)aPacket;
            mInputQueue.put(new Red5Message(type, dataPacket));
          }
        } catch (InterruptedException ex)
        {
        } finally {
        }
      }

    };
    mOutputHandler = new IRTMPEventIOHandler()
    {
      /**
       * Reading not supported on this handler.
       * @return null
       */
      public Red5Message read() throws InterruptedException
      {
        return null;
      }

      public void write(Red5Message aMsg) throws InterruptedException
      {
        try {

          IRTMPEvent event = aMsg.getData();
          if (event != null)
          {
            // we've live.  Tell Red5 that
            // this is only in the xuggler_timestamp_branch for now, so we leave commented out
            //event.setSourceType((byte)1);

            mOutputStream.dispatchEvent(event);
            event.release();
          }
        } finally {
        }
      }
    };

  }

  /**
   * Is the main loop running?
   * 
   * @see #run()
   * @return true if the loop is running, false otherwise.
   */
  public boolean isRunning()
  {
    return mIsRunning;
  }

  /**
   * Stop the {@link Transcoder} loop if it's running
   * on a separate thread.
   * <p>
   * It does this by sending a
   * {@link Red5Message} for the end of stream 
   * 
   * to the {@link Transcoder} and allowing it to
   * exit gracefully.
   * </p>
   * @see #run()
   */
  public void stop()
  {
    try
    {
      mInputQueue.put(new Red5Message(Red5Message.Type.END_STREAM, null));
    }
    catch (InterruptedException e)
    {
    }
    mKeepRunning = false;
  }

  /**
   * Open up all input and ouput containers (files)
   * and being transcoding.
   * <p>
   * The {@link Transcoder} requires its own thread to
   * do work on, and callers are responsible for
   * allocating the {@link Thread}.
   * </p>
   * <p>
   * This method does not return unless another thread
   * calls {@link Transcoder#stop()}, or it reaches
   * the end of a Red5 stream.  It is meant to
   * be passed as the {@link Runnable#run()} method
   * for a thread.
   * </p>
   */
  public void run()
  {
    try
    {
      openContainer();
      transcode();
    }
    catch (Throwable e)
    {
    }
    finally
    {
      closeContainer();
    }
  }

  private void transcode()
  {
    int retval = -1;
    synchronized(this)
    {
      mIsRunning = true;
      notifyAll();
    }
  
    IPacket iPacket = IPacket.make();
  
    while (mKeepRunning) {
  
      try {
        try {
          retval = mInContainer.readNextPacket(iPacket);
        } finally {
        }
        if (retval <0)
        {
          mKeepRunning = false;
          break;
        }
        IPacket decodePacket = iPacket;
        if (mPacketListener != null)
        {
          try {
            decodePacket = mPacketListener.preDecode(iPacket);
            if (decodePacket == null)
              decodePacket = iPacket;
          } finally {
          }
        }
        openInputCoders(decodePacket); // reopen the input coders if we need to
        int i = decodePacket.getStreamIndex();
        if (i == mAudioStreamId) {
          if (mInAudioCoder == null)
            throw new RuntimeException("audio coder not set up");
            
          if (mOutputInfo.hasAudio())
            decodeAudio(decodePacket);
        } else if (i == mVideoStreamId) {
          if (mInVideoCoder == null)
            throw new RuntimeException("video coder not set up");

          if (mOutputInfo.hasVideo())
            decodeVideo(decodePacket);
        } else {
        }
      } finally {
      }
    }
  }

  private void openContainer()
  {
    try {
      // set out thread name
      String threadName = "Transcoder["+mInputStream.getPublishedName()+"]";
      Thread.currentThread().setName(threadName);
      int retval = -1;
  
      // First let's setup our input URL
      {
        // Register a new listener; should hopefully start getting audio packets immediately
        mInputStream.addStreamListener(mInputListener);
  
        // Tell AAFFMPEG about our new input URL; we use the unique Red5 names for the url
        mInputURL = Red5HandlerFactory.DEFAULT_PROTOCOL+":"+mInputStream.getName();
        ISimpleMediaFile inputInfo = new SimpleMediaFile();
        inputInfo.setURL(mInputURL);
        mFactory.registerStream(mInputQueue, inputInfo);
  
        mInContainer = IContainer.make();
        // NOTE: This will block until we get the later of the first audio if it has audio, or first video
        // if it has video
        IContainerFormat inFormat = IContainerFormat.make();
        inFormat.setInputFormat("flv"); // set the input format to avoid FFMPEG probing
        retval = mInContainer.open(mInputURL, IContainer.Type.READ, inFormat, true, false);
        if (retval < 0)
        {
          throw new RuntimeException("Could not open input: " + mInputURL);
        }
      }
      // Now, let's first set up our output URL
      {
        // Tell AAFFMPEG about out output URL
        mOutputURL = Red5HandlerFactory.DEFAULT_PROTOCOL+":"+mOutputStream.getName();
        mOutputInfo.setURL(mOutputURL);
        // For the output URL, every time we get a packet we just dispatch it to
        // a stream; you also use a Red5StreamingQueue here if you wanted to
        // have another thread deal with broadcasting.
        mFactory.registerStream(mOutputHandler, mOutputInfo);
  
        mOutContainer = IContainer.make();
        IContainerFormat outFormat = IContainerFormat.make();
        outFormat.setOutputFormat("flv", mOutputURL, null);
        retval = mOutContainer.open(mOutputURL, IContainer.Type.WRITE, outFormat);
        if (retval < 0)
          throw new RuntimeException("could not open output: "+mOutputURL);
  
        if (mOutputInfo.hasAudio())
        {
          // Add an audio stream
          IStream outStream = mOutContainer.addNewStream(0);
          if (outStream == null)
            throw new RuntimeException("could not add audio stream to output: "+mOutputURL);
          IStreamCoder outCoder = outStream.getStreamCoder();
          ICodec.ID outCodecId = mOutputInfo.getAudioCodec();
          ICodec outCodec = ICodec.findEncodingCodec(outCodecId);
          if (outCodec == null)
          {
            throw new RuntimeException("Could not encode using the codec: "+mOutputInfo.getAudioCodec());
          }
          outCoder.setCodec(outCodec);
          outCoder.setBitRate(mOutputInfo.getAudioBitRate());
            
          outCoder.setSampleRate(mOutputInfo.getAudioSampleRate());
          
          outCoder.setChannels(mOutputInfo.getAudioChannels());
          
          outCoder.open();
          // if we get here w/o an exception, record the coder
          mOutAudioCoder = outCoder;
        }
        if (mOutputInfo.hasVideo())
        {
          // Add an audio stream
          IStream outStream = mOutContainer.addNewStream(1);
          if (outStream == null)
            throw new RuntimeException("could not add video stream to output: "+mOutputURL);
          IStreamCoder outCoder = outStream.getStreamCoder();
          ICodec.ID outCodecId = mOutputInfo.getVideoCodec();
          ICodec outCodec = ICodec.findEncodingCodec(outCodecId);
          if (outCodec == null)
          {
            throw new RuntimeException("Could not encode using the codec: "+mOutputInfo.getAudioCodec());
          }
          outCoder.setCodec(outCodec);
          outCoder.setWidth(mOutputInfo.getVideoWidth());
          outCoder.setHeight(mOutputInfo.getVideoHeight());
          outCoder.setPixelType(mOutputInfo.getVideoPixelFormat());
          outCoder.setGlobalQuality(mOutputInfo.getVideoGlobalQuality());
          outCoder.setBitRate(mOutputInfo.getVideoBitRate());
          outCoder.setNumPicturesInGroupOfPictures(mOutputInfo.getVideoNumPicturesInGroupOfPictures());
          outCoder.setFlag(IStreamCoder.Flags.FLAG_QSCALE, true);

          if (mOutputInfo.getVideoTimeBase() != null)
            outCoder.setTimeBase(mOutputInfo.getVideoTimeBase());
          else
            outCoder.setTimeBase(IRational.make(1, 1000)); // default to FLV

          outCoder.open();
          // if we get here w/o an exception, record the coder
          mOutVideoCoder = outCoder;
        }
      }
      retval = mOutContainer.writeHeader();
      if (retval < 0)
      {
        throw new RuntimeException("could not write header for output");
      }
    } finally {
    }
  
  }

  private void openInputCoders(IPacket packet)
  {
    {
      IStreamCoder audioCoder = null;
      IStreamCoder videoCoder = null;
      if (mAudioStreamId == -1 || mVideoStreamId == -1)
      {
        int numStreams = mInContainer.getNumStreams();

        for(int i  = 0; i < numStreams; i++)
        {
          IStream stream = mInContainer.getStream(i);
          if (stream != null)
          {
            IStreamCoder coder = stream.getStreamCoder();
            if (coder != null)
            {
              if (coder.getCodecType()==ICodec.Type.CODEC_TYPE_AUDIO // if audio
                  && mAudioStreamId == -1 // and we haven't already initialized
                  && packet.getStreamIndex() == i // and this packet is also audio
                  )
              {
                if (coder.getCodec() != null)
                {
                  audioCoder = coder;
                  mAudioStreamId = i;
                } else {
                  throw new RuntimeException("Could not find codec for audio stream");
                }
              }
              if (coder.getCodecType()==ICodec.Type.CODEC_TYPE_VIDEO // if video
                  && mVideoStreamId == -1 // and we haven't already initialized
                  && packet.getStreamIndex() == i // and this packet is also video
                  )
              {
                if (coder.getCodec() != null)
                {
                  videoCoder = coder;
                  mVideoStreamId = i;
                } else {
                  throw new RuntimeException("Could not find codec for video stream");
                }
              }
            }
          }
        }
      }
      if (mAudioStreamId != -1 && mInAudioCoder == null)
      {
        if (audioCoder.open() < 0)
          throw new RuntimeException("could not open audio coder for stream: " + mAudioStreamId);
        mInAudioCoder = audioCoder;
      }
      
      if (mVideoStreamId != -1 &&  mInVideoCoder == null)
      {
        if (videoCoder.open() < 0)
          throw new RuntimeException("could not open video coder for stream: " + mVideoStreamId);
        mInVideoCoder = videoCoder;
      }
    }
  }

  private void openVideoResampler(IVideoPicture picture)
  {
    if (mVideoResampler == null && mOutVideoCoder != null)
    {
      if (picture.getWidth() <= 0 || picture.getHeight()<=0)
        throw new RuntimeException("frame has no data in it so cannot resample");

      // We set up our resampler.
      if (mOutVideoCoder.getPixelType() != picture.getPixelType() ||
          mOutVideoCoder.getWidth() != picture.getWidth() ||
          mOutVideoCoder.getHeight() != picture.getHeight())
      {
        mVideoResampler = IVideoResampler.make(
            mOutVideoCoder.getWidth(),
            mOutVideoCoder.getHeight(),
            mOutVideoCoder.getPixelType(),
            picture.getWidth(),
            picture.getHeight(),
            picture.getPixelType());
        if (mVideoResampler == null)
        {
          throw new RuntimeException("needed to resample video but couldn't allocate a resampler; you need the GPL version of AAFFMPEG installed?");
        }
      }
    }
  }

  private void openAudioResampler(IAudioSamples samples)
  {
    if (mAudioResampler == null && mOutAudioCoder != null)
    {
      if (mOutAudioCoder.getSampleRate() != samples.getSampleRate() ||
          mOutAudioCoder.getChannels() != samples.getChannels())
      {
        mAudioResampler = IAudioResampler.make(
            mOutAudioCoder.getChannels(),
            samples.getChannels(),
            mOutAudioCoder.getSampleRate(),
            samples.getSampleRate());
        if (mAudioResampler == null)
        {
          throw new RuntimeException("needed to resample audio but couldn't allocate a resampler");
        }
      }
      // and we write the output header
    }
  }

  private void closeContainer()
  {
    try
    {
      try
      {
        mInputStream.removeStreamListener(mInputListener);
        if (mOutContainer != null)
          mOutContainer.writeTrailer();
        if (mOutAudioCoder != null)
          mOutAudioCoder.close();
        mOutAudioCoder = null;
        if (mInAudioCoder != null)
          mInAudioCoder.close();
        mInAudioCoder = null;
        if (mOutVideoCoder != null)
          mOutVideoCoder.close();
        mOutVideoCoder = null;
        if (mInVideoCoder != null)
          mInVideoCoder.close();
        mInVideoCoder = null;
        if (mOutContainer != null)
          mOutContainer.close();
        mOutContainer = null;
      }
      finally
      {
        synchronized(this)
        {
          mIsRunning = false;
          notifyAll();
        }
      }
    } finally {
    }
  
  }

  private void writePacket(IPacket oPacket)
  {
    int retval;
    IPacket encodedPacket = oPacket;
    if (mPacketListener != null)
    {
      try {
        encodedPacket = mPacketListener.postEncode(oPacket);
        if (encodedPacket == null)
          encodedPacket = oPacket;
      } finally {
      }
    }
    // Don't force interleaving of data
    try {
      retval = mOutContainer.writePacket(encodedPacket,
          false);
    } finally {
    }
  
    if (retval < 0) {
      throw new RuntimeException(
      "could not write output packet");
    }
  }

  private void decodeVideo(IPacket decodePacket)
  {
    int retval = -1;
    // Note that we don't specify the input width and height; the StreamCoder will fill that
    // in when it decodes
    IVideoPicture inPicture = IVideoPicture.make(mInVideoCoder.getPixelType(), mInVideoCoder.getWidth(), mInVideoCoder.getHeight());
    // resampled video
    IVideoPicture reSample = null;
  
    int offset = 0;
  
    while (offset < decodePacket.getSize()) {
      try{
        retval = mInVideoCoder.decodeVideo(inPicture, decodePacket, offset);
      } finally {
      }
      if (retval <= 0) {
        return;
      }
      offset += retval;
      
      IVideoPicture postDecode = inPicture;
      if (mVideoPictureListener != null)
      {
        try {
          postDecode = mVideoPictureListener.postDecode(inPicture);
          if (postDecode == null)
            postDecode = inPicture;
        } finally {
        }
      }
      
      if (postDecode.isComplete())
      {
        reSample = resampleVideo(postDecode);
      }
      else
      {
        reSample = postDecode;
      }
      if (reSample.isComplete())
      {
        encodeVideo(reSample);
      }
    }
  }

  private void encodeVideo(IVideoPicture picture)
  {
    int retval;
    IPacket oPacket = IPacket.make();

    /**
     * NOTE: At this point reSamples contains the actual unencoded raw samples.
     * 
     * The next step does an encoding, but you PROBABLY don't need to do that.
     * Instead, you could copy the reSamples.getSamples().getData(...) bytes
     * into your own structure and hand them off, but for now, we'll
     * try re-encoding as FLV with PCM embedded.
     */
    IVideoPicture preEncode= picture;
    if (mVideoPictureListener != null)
    {
      try {
        preEncode = mVideoPictureListener.preEncode(picture);
        if (preEncode == null)
          preEncode = picture;
      } finally {
      }
    }

    int numBytesConsumed = 0;
    if (preEncode.isComplete()) {
      try {
        retval = mOutVideoCoder.encodeVideo(oPacket, preEncode, 0);
      } finally {
      }
      if (retval <= 0) {
        // If we fail to encode, complain loudly but still keep going
      } else {
        numBytesConsumed += retval;
      }
      if (oPacket.isComplete()) {
        writePacket(oPacket);
      }
    }
  }

  private IVideoPicture resampleVideo(IVideoPicture picture)
  {
    IVideoPicture reSample;
    
    openVideoResampler(picture);

    if (mVideoResampler != null) {

      IVideoPicture outPicture = IVideoPicture.make(mOutVideoCoder.getPixelType(), mOutVideoCoder.getWidth(), mOutVideoCoder.getHeight());

      IVideoPicture preResample= picture;
      if (mVideoPictureListener != null)
      {
        try {
          preResample = mVideoPictureListener.preResample(picture);
          if (preResample == null)
            preResample = picture;
        } finally {
        }
      }

      int retval = -1;
      try{
        retval = mVideoResampler.resample(outPicture, preResample);
      } finally {
      }
      if (retval < 0)
        throw new RuntimeException("could not resample video");
      IVideoPicture postResample= outPicture;
      if (mVideoPictureListener != null)
      {
        try {
          postResample = mVideoPictureListener.postResample(outPicture);
          if (postResample == null)
            postResample = outPicture;
        } finally {
        }
      }

      reSample = postResample;
    } else {
      reSample = picture;
    }
    return reSample;
  }

  private void decodeAudio(IPacket decodePacket)
  {
    int retval = -1;
    IAudioSamples inSamples = IAudioSamples.make(1024, mInAudioCoder.getChannels());
    // resampled audio
    IAudioSamples reSamples = null;
  
    int offset = 0;
  
    while (offset < decodePacket.getSize()) {
      try{
        retval = mInAudioCoder.decodeAudio(inSamples, decodePacket, offset);
      } finally {
      }
      if (retval <= 0) {
        throw new RuntimeException("could not decode audio");
      }
      offset += retval;
      
      IAudioSamples postDecode = inSamples;
      if (mSamplesListener != null)
      {
        try {
          postDecode = mSamplesListener.postDecode(inSamples);
          if (postDecode == null)
            postDecode = inSamples;
        } finally {
        }
      }
      
      if (postDecode.isComplete())
      {
        reSamples = resampleAudio(postDecode);
      } else
      {
        reSamples = postDecode;
      }
      
      if (reSamples.isComplete())
      {
        encodeAudio(reSamples);
      }
    }    
  }

  private void encodeAudio(IAudioSamples samples)
  {
    int retval;
    IPacket oPacket = IPacket.make();

    /**
     * NOTE: At this point reSamples contains the actual unencoded raw samples.
     * 
     * The next step does an encoding, but you PROBABLY don't need to do that.
     * Instead, you could copy the reSamples.getSamples().getData(...) bytes
     * into your own structure and hand them off, but for now, we'll
     * try re-encoding as FLV with PCM embedded.
     */
    IAudioSamples preEncode= samples;
    if (mSamplesListener != null)
    {
      try {
        preEncode = mSamplesListener.preEncode(samples);
        if (preEncode == null)
          preEncode = samples;
      } finally {
      }
    }

    int numSamplesConsumed = 0;
    while (numSamplesConsumed < preEncode.getNumSamples()) {
      try {
        retval = mOutAudioCoder.encodeAudio(oPacket, preEncode,
            numSamplesConsumed);
      } finally {
      }
      if (retval <= 0) {
        // If we fail to encode, complain loudly but still keep going
        // and break the loop since these samples are now suspect
        break;
      }

      numSamplesConsumed += retval;

      if (oPacket.isComplete()) {
        writePacket(oPacket);
      }
    }
  }

  private IAudioSamples resampleAudio(IAudioSamples samples)
  {
    IAudioSamples reSamples;
    
    openAudioResampler(samples);
    
    IAudioSamples outSamples = IAudioSamples.make(1024, mOutAudioCoder.getChannels());

    if (mAudioResampler != null && samples.getNumSamples() > 0) {
 
      IAudioSamples preResample= samples;
      if (mSamplesListener != null)
      {
        try {
          preResample = mSamplesListener.preResample(samples);
          if (preResample == null)
            preResample = samples;
        } finally {
        }
      }
      
      int retval = -1;
      try{
        retval = mAudioResampler.resample(outSamples, preResample,
            preResample.getNumSamples());
      } finally {
      }
      if (retval < 0)
        throw new RuntimeException("could not resample audio");
      
      IAudioSamples postResample= outSamples;
      if (mSamplesListener != null)
      {
        try {
          postResample = mSamplesListener.postResample(outSamples);
          if (postResample == null)
            postResample = outSamples;
        } finally {
        }
      }
 
      reSamples = postResample;
    } else {
      reSamples = samples;
    }
    return reSamples;
  }
}
